//----------------------------------------------------------------------------
//
// LFSR functions with x86 optimizations, by Scott Duplichan
// tested with Microsoft and mingw C compilers
//
//----------------------------------------------------------------------------
#include "project.h"
//----------------------------------------------------------------------------
// bit shift left a 128-bit value using xmm registers
//          __m128i *data - data to shift
//          int count     - number of bits to shift
// return:  __m128i       - carry out bit(s)

static __m128i bitShiftLeft128xmm (__m128i *data, int count)
   {
   __m128i innerCarry, carryOut;

   innerCarry = _mm_srli_epi64 (*data, 64 - count);      // carry outs in bit 0 of each qword
   carryOut   = _mm_shuffle_epi32 (innerCarry, 0xFE);    // upper carry in xmm bit 0, others zero
   innerCarry = _mm_shuffle_epi32 (innerCarry, 0xCF);    // lower carry in xmm bit 64, others zero
   *data = _mm_slli_epi64 (*data, count);                // shift all qwords left
   *data = _mm_or_si128 (*data, innerCarry);             // propagate carry out from low qword
   return carryOut;
   }

//----------------------------------------------------------------------------
// bit shift left a 256-bit value using xmm registers
//          __m128i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m128i       - carry out bit(s)

static __m128i bitShiftLeft256xmm (__m128i *data, int count)
   {
   __m128i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft128xmm (&data [0], count);
   carryOut1 = bitShiftLeft128xmm (&data [1], count);
   data [1] = _mm_or_si128 (data [1], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
// bit shift left a 512-bit value using xmm registers
//          __m128i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m128i       - carry out bit(s)

static __m128i bitShiftLeft512xmm (__m128i *data, int count)
   {
   __m128i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft256xmm (&data [0], count);
   carryOut1 = bitShiftLeft256xmm (&data [2], count);
   data [2] = _mm_or_si128 (data [2], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
// bit shift left a 1024-bit value using xmm registers
//          __m128i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m128i       - carry out bit(s)

static __m128i bitShiftLeft1024xmm (__m128i *data, int count)
   {
   __m128i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft512xmm (&data [0], count);
   carryOut1 = bitShiftLeft512xmm (&data [4], count);
   data [4] = _mm_or_si128 (data [4], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
// bit shift left a 2048-bit value using xmm registers
//          __m128i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m128i       - carry out bit(s)

static __m128i bitShiftLeft2048xmm (__m128i *data, int count)
   {
   __m128i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft1024xmm (&data [0], count);
   carryOut1 = bitShiftLeft1024xmm (&data [8], count);
   data [8] = _mm_or_si128 (data [8], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
// bit shift left a 4096-bit value using xmm registers
//          __m128i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m128i       - carry out bit(s)

static __m128i bitShiftLeft4096xmm (__m128i *data, int count)
   {
   __m128i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft2048xmm (&data [0], count);
   carryOut1 = bitShiftLeft2048xmm (&data [16], count);
   data [16] = _mm_or_si128 (data [16], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
//
// lfsr32xmm - left shift galois type lfsr for 32-bit data, xmm implementation
//

static __m128i lfsr32xmm (__m128i data, __m128i mask)
   {
   __m128i maskOrZero;

   // expand data bit 31 into mask 0-63
   maskOrZero = _mm_srai_epi32 (data, 31);            // lower 32 bits all zeros or all ones

   // zero the mask if data bit 31 is zero
   maskOrZero = _mm_and_si128 (maskOrZero, mask);

   // shift the whole data left by one bit
   data = _mm_slli_epi32 (data, 1);

   // xor in mask if bit 31 was set
   return _mm_xor_si128 (data, maskOrZero);
   }

//----------------------------------------------------------------------------
//
// lfsr64xmm - left shift galois type lfsr for 64-bit data, xmm implementation
//

static __m128i lfsr64xmm (__m128i data, __m128i mask)
   {
   __m128i maskOrZero;

   // expand data bit 63 into mask 0-63
   maskOrZero = _mm_shuffle_epi32 (data, 0xE5);       // copy high dword to low dword
   maskOrZero = _mm_srai_epi32 (maskOrZero, 31);      // lower 64 bits all zeros or all ones

   // zero the mask if data bit 63 is zero
   maskOrZero = _mm_and_si128 (maskOrZero, mask);

   // shift the whole data left by one bit
   data = _mm_slli_epi64 (data, 1);

   // xor in mask if bit 63 was set
   return _mm_xor_si128 (data, maskOrZero);
   }

//----------------------------------------------------------------------------
//
// lfsr128xmm - left shift galois type lfsr for 128-bit data, xmm implementation
//

static __m128i lfsr128xmm (__m128i data, __m128i mask)
   {
   __m128i maskOrZero;

   // expand data bit 127 into mask 0-127
   maskOrZero = _mm_shuffle_epi32 (data, 0xFF);       // copy high dword to all 4 dwords
   maskOrZero = _mm_srai_epi32 (maskOrZero, 31);      // all zeros or all ones

   // zero the mask if data bit 127 is zero
   maskOrZero = _mm_and_si128 (maskOrZero, mask);

   // shift the whole data left by one bit
   bitShiftLeft128xmm (&data, 1);

   // xor in mask if bit 127 was set
   return _mm_xor_si128 (data, maskOrZero);
   }

//----------------------------------------------------------------------------
//
// lfsr256xmm - left shift galois type lfsr for 256-bit data, xmm implementation
//

static void lfsr256xmm (__m128i *data, __m128i *mask)
   {
   __m128i maskOrZero [2], maskOrZ;

   // expand data bit 255 into mask 0-255
   maskOrZ = _mm_shuffle_epi32 (data [1], 0xFF);      // copy high dword to all 4 dwords
   maskOrZ = _mm_srai_epi32 (maskOrZ, 31);            // all zeros or all ones

   // zero the mask if data bit 255 is zero
   maskOrZero [0] = _mm_and_si128 (maskOrZ, mask [0]);
   maskOrZero [1] = _mm_and_si128 (maskOrZ, mask [1]);

   // shift the whole data left by one bit
   bitShiftLeft256xmm (data, 1);

   // xor in mask if bit 255 was set
   data [0] = _mm_xor_si128 (data [0], maskOrZero [0]);
   data [1] = _mm_xor_si128 (data [1], maskOrZero [1]);
   }

//----------------------------------------------------------------------------
//
// lfsr512xmm - left shift galois type lfsr for 512-bit data, xmm implementation
//

static void lfsr512xmm (__m128i *data, __m128i *mask)
   {
   __m128i maskOrZero [4], maskOrZ;

   // expand data bit 511 into mask 0-511
   maskOrZ = _mm_shuffle_epi32 (data [3], 0xFF);      // copy high dword to all 4 dwords
   maskOrZ = _mm_srai_epi32 (maskOrZ, 31);            // all zeros or all ones

   // zero the mask if data bit 511 is zero
   maskOrZero [0] = _mm_and_si128 (maskOrZ, mask [0]);
   maskOrZero [1] = _mm_and_si128 (maskOrZ, mask [1]);
   maskOrZero [2] = _mm_and_si128 (maskOrZ, mask [2]);
   maskOrZero [3] = _mm_and_si128 (maskOrZ, mask [3]);

   // shift the whole data left by one bit
   bitShiftLeft512xmm (data, 1);

   // xor in mask if bit 511 was set
   data [0] = _mm_xor_si128 (data [0], maskOrZero [0]);
   data [1] = _mm_xor_si128 (data [1], maskOrZero [1]);
   data [2] = _mm_xor_si128 (data [2], maskOrZero [2]);
   data [3] = _mm_xor_si128 (data [3], maskOrZero [3]);
   }

//----------------------------------------------------------------------------
//
// lfsr1024xmm - left shift galois type lfsr for 1024-bit data, xmm implementation
//

static void lfsr1024xmm (__m128i *data, __m128i *mask)
   {
   __m128i maskOrZero [8], maskOrZ;

   // expand data bit 1023 into mask 0-1023
   maskOrZ = _mm_shuffle_epi32 (data [7], 0xFF);      // copy high dword to all 4 dwords
   maskOrZ = _mm_srai_epi32 (maskOrZ, 31);            // all zeros or all ones

   // zero the mask if data bit 1023 is zero
   maskOrZero [0] = _mm_and_si128 (maskOrZ, mask [0]);
   maskOrZero [1] = _mm_and_si128 (maskOrZ, mask [1]);
   maskOrZero [2] = _mm_and_si128 (maskOrZ, mask [2]);
   maskOrZero [3] = _mm_and_si128 (maskOrZ, mask [3]);
   maskOrZero [4] = _mm_and_si128 (maskOrZ, mask [4]);
   maskOrZero [5] = _mm_and_si128 (maskOrZ, mask [5]);
   maskOrZero [6] = _mm_and_si128 (maskOrZ, mask [6]);
   maskOrZero [7] = _mm_and_si128 (maskOrZ, mask [7]);

   // shift the whole data left by one bit
   bitShiftLeft1024xmm (data, 1);

   // xor in mask if bit 1023 was set
   data [0] = _mm_xor_si128 (data [0], maskOrZero [0]);
   data [1] = _mm_xor_si128 (data [1], maskOrZero [1]);
   data [2] = _mm_xor_si128 (data [2], maskOrZero [2]);
   data [3] = _mm_xor_si128 (data [3], maskOrZero [3]);
   data [4] = _mm_xor_si128 (data [4], maskOrZero [4]);
   data [5] = _mm_xor_si128 (data [5], maskOrZero [5]);
   data [6] = _mm_xor_si128 (data [6], maskOrZero [6]);
   data [7] = _mm_xor_si128 (data [7], maskOrZero [7]);
   }

//----------------------------------------------------------------------------
//
// lfsr2048xmm - left shift galois type lfsr for 2048-bit data, xmm implementation
//

static void lfsr2048xmm (__m128i *data, __m128i *mask)
   {
   __m128i maskOrZero [16], maskOrZ;

   // expand data bit 2047 into mask 0-2047
   maskOrZ = _mm_shuffle_epi32 (data [15], 0xFF);     // copy high dword to all 4 dwords
   maskOrZ = _mm_srai_epi32 (maskOrZ, 31);            // all zeros or all ones

   // zero the mask if data bit 2047 is zero
   maskOrZero [ 0] = _mm_and_si128 (maskOrZ, mask [ 0]);
   maskOrZero [ 1] = _mm_and_si128 (maskOrZ, mask [ 1]);
   maskOrZero [ 2] = _mm_and_si128 (maskOrZ, mask [ 2]);
   maskOrZero [ 3] = _mm_and_si128 (maskOrZ, mask [ 3]);
   maskOrZero [ 4] = _mm_and_si128 (maskOrZ, mask [ 4]);
   maskOrZero [ 5] = _mm_and_si128 (maskOrZ, mask [ 5]);
   maskOrZero [ 6] = _mm_and_si128 (maskOrZ, mask [ 6]);
   maskOrZero [ 7] = _mm_and_si128 (maskOrZ, mask [ 7]);
   maskOrZero [ 8] = _mm_and_si128 (maskOrZ, mask [ 8]);
   maskOrZero [ 9] = _mm_and_si128 (maskOrZ, mask [ 9]);
   maskOrZero [10] = _mm_and_si128 (maskOrZ, mask [10]);
   maskOrZero [11] = _mm_and_si128 (maskOrZ, mask [11]);
   maskOrZero [12] = _mm_and_si128 (maskOrZ, mask [12]);
   maskOrZero [13] = _mm_and_si128 (maskOrZ, mask [13]);
   maskOrZero [14] = _mm_and_si128 (maskOrZ, mask [14]);
   maskOrZero [15] = _mm_and_si128 (maskOrZ, mask [15]);

   // shift the whole data left by one bit
   bitShiftLeft2048xmm (data, 1);

   // xor in mask if bit 2047 was set
   data [ 0] = _mm_xor_si128 (data [ 0], maskOrZero [ 0]);
   data [ 1] = _mm_xor_si128 (data [ 1], maskOrZero [ 1]);
   data [ 2] = _mm_xor_si128 (data [ 2], maskOrZero [ 2]);
   data [ 3] = _mm_xor_si128 (data [ 3], maskOrZero [ 3]);
   data [ 4] = _mm_xor_si128 (data [ 4], maskOrZero [ 4]);
   data [ 5] = _mm_xor_si128 (data [ 5], maskOrZero [ 5]);
   data [ 6] = _mm_xor_si128 (data [ 6], maskOrZero [ 6]);
   data [ 7] = _mm_xor_si128 (data [ 7], maskOrZero [ 7]);
   data [ 8] = _mm_xor_si128 (data [ 8], maskOrZero [ 8]);
   data [ 9] = _mm_xor_si128 (data [ 9], maskOrZero [ 9]);
   data [10] = _mm_xor_si128 (data [10], maskOrZero [10]);
   data [11] = _mm_xor_si128 (data [11], maskOrZero [11]);
   data [12] = _mm_xor_si128 (data [12], maskOrZero [12]);
   data [13] = _mm_xor_si128 (data [13], maskOrZero [13]);
   data [14] = _mm_xor_si128 (data [14], maskOrZero [14]);
   data [15] = _mm_xor_si128 (data [15], maskOrZero [15]);
   }

//----------------------------------------------------------------------------
//
// lfsr4096xmm - left shift galois type lfsr for 4096-bit data, xmm implementation
//

static void lfsr4096xmm (__m128i *data, __m128i *mask)
   {
   __m128i maskOrZero [32], maskOrZ;

   // expand data bit 4095 into mask 0-4095
   maskOrZ = _mm_shuffle_epi32 (data [31], 0xFF);     // copy high dword to all 4 dwords
   maskOrZ = _mm_srai_epi32 (maskOrZ, 31);            // all zeros or all ones

   // zero the mask if data bit 4095 is zero
   maskOrZero [ 0] = _mm_and_si128 (maskOrZ, mask [ 0]);
   maskOrZero [ 1] = _mm_and_si128 (maskOrZ, mask [ 1]);
   maskOrZero [ 2] = _mm_and_si128 (maskOrZ, mask [ 2]);
   maskOrZero [ 3] = _mm_and_si128 (maskOrZ, mask [ 3]);
   maskOrZero [ 4] = _mm_and_si128 (maskOrZ, mask [ 4]);
   maskOrZero [ 5] = _mm_and_si128 (maskOrZ, mask [ 5]);
   maskOrZero [ 6] = _mm_and_si128 (maskOrZ, mask [ 6]);
   maskOrZero [ 7] = _mm_and_si128 (maskOrZ, mask [ 7]);
   maskOrZero [ 8] = _mm_and_si128 (maskOrZ, mask [ 8]);
   maskOrZero [ 9] = _mm_and_si128 (maskOrZ, mask [ 9]);
   maskOrZero [10] = _mm_and_si128 (maskOrZ, mask [10]);
   maskOrZero [11] = _mm_and_si128 (maskOrZ, mask [11]);
   maskOrZero [12] = _mm_and_si128 (maskOrZ, mask [12]);
   maskOrZero [13] = _mm_and_si128 (maskOrZ, mask [13]);
   maskOrZero [14] = _mm_and_si128 (maskOrZ, mask [14]);
   maskOrZero [15] = _mm_and_si128 (maskOrZ, mask [15]);
   maskOrZero [16] = _mm_and_si128 (maskOrZ, mask [16]);
   maskOrZero [17] = _mm_and_si128 (maskOrZ, mask [17]);
   maskOrZero [18] = _mm_and_si128 (maskOrZ, mask [18]);
   maskOrZero [19] = _mm_and_si128 (maskOrZ, mask [19]);
   maskOrZero [20] = _mm_and_si128 (maskOrZ, mask [20]);
   maskOrZero [21] = _mm_and_si128 (maskOrZ, mask [21]);
   maskOrZero [22] = _mm_and_si128 (maskOrZ, mask [22]);
   maskOrZero [23] = _mm_and_si128 (maskOrZ, mask [23]);
   maskOrZero [24] = _mm_and_si128 (maskOrZ, mask [24]);
   maskOrZero [25] = _mm_and_si128 (maskOrZ, mask [25]);
   maskOrZero [26] = _mm_and_si128 (maskOrZ, mask [26]);
   maskOrZero [27] = _mm_and_si128 (maskOrZ, mask [27]);
   maskOrZero [28] = _mm_and_si128 (maskOrZ, mask [28]);
   maskOrZero [29] = _mm_and_si128 (maskOrZ, mask [29]);
   maskOrZero [30] = _mm_and_si128 (maskOrZ, mask [30]);
   maskOrZero [31] = _mm_and_si128 (maskOrZ, mask [31]);

   // shift the whole data left by one bit
   bitShiftLeft4096xmm (data, 1);

   // xor in mask if bit 4095 was set
   data [ 0] = _mm_xor_si128 (data [ 0], maskOrZero [ 0]);
   data [ 1] = _mm_xor_si128 (data [ 1], maskOrZero [ 1]);
   data [ 2] = _mm_xor_si128 (data [ 2], maskOrZero [ 2]);
   data [ 3] = _mm_xor_si128 (data [ 3], maskOrZero [ 3]);
   data [ 4] = _mm_xor_si128 (data [ 4], maskOrZero [ 4]);
   data [ 5] = _mm_xor_si128 (data [ 5], maskOrZero [ 5]);
   data [ 6] = _mm_xor_si128 (data [ 6], maskOrZero [ 6]);
   data [ 7] = _mm_xor_si128 (data [ 7], maskOrZero [ 7]);
   data [ 8] = _mm_xor_si128 (data [ 8], maskOrZero [ 8]);
   data [ 9] = _mm_xor_si128 (data [ 9], maskOrZero [ 9]);
   data [10] = _mm_xor_si128 (data [10], maskOrZero [10]);
   data [11] = _mm_xor_si128 (data [11], maskOrZero [11]);
   data [12] = _mm_xor_si128 (data [12], maskOrZero [12]);
   data [13] = _mm_xor_si128 (data [13], maskOrZero [13]);
   data [14] = _mm_xor_si128 (data [14], maskOrZero [14]);
   data [15] = _mm_xor_si128 (data [15], maskOrZero [15]);
   data [16] = _mm_xor_si128 (data [16], maskOrZero [16]);
   data [17] = _mm_xor_si128 (data [17], maskOrZero [17]);
   data [18] = _mm_xor_si128 (data [18], maskOrZero [18]);
   data [19] = _mm_xor_si128 (data [19], maskOrZero [19]);
   data [20] = _mm_xor_si128 (data [20], maskOrZero [20]);
   data [21] = _mm_xor_si128 (data [21], maskOrZero [21]);
   data [22] = _mm_xor_si128 (data [22], maskOrZero [22]);
   data [23] = _mm_xor_si128 (data [23], maskOrZero [23]);
   data [24] = _mm_xor_si128 (data [24], maskOrZero [24]);
   data [25] = _mm_xor_si128 (data [25], maskOrZero [25]);
   data [26] = _mm_xor_si128 (data [26], maskOrZero [26]);
   data [27] = _mm_xor_si128 (data [27], maskOrZero [27]);
   data [28] = _mm_xor_si128 (data [28], maskOrZero [28]);
   data [29] = _mm_xor_si128 (data [29], maskOrZero [29]);
   data [30] = _mm_xor_si128 (data [30], maskOrZero [30]);
   data [31] = _mm_xor_si128 (data [31], maskOrZero [31]);
   }

//===========================================================================
//
// measure execution time and verify proper operation for each lfsr function
//

void lfsrXmm (TEST_DATA *testData, int loops, uint64_t cpuSpeed)
   {
   int index;
   uint64_t startCount;
   __m128i pattern32, mask32;
   __m128i pattern64, mask64;
   __m128i pattern128, mask128;
   static __m128i  pattern256 [2], pattern512 [4], pattern1024 [8], pattern2048 [16], pattern4096 [32];
   static __m128i  mask256    [2], mask512    [4], mask1024    [8], mask2048    [16], mask4096    [32];

   memcpy (&pattern32 , testData->initial, sizeof pattern32  );
   memcpy (&pattern64 , testData->initial, sizeof pattern64  );
   memcpy (&pattern128, testData->initial, sizeof pattern128 );
   memcpy (pattern256 , testData->initial, sizeof pattern256 );
   memcpy (pattern512 , testData->initial, sizeof pattern512 );
   memcpy (pattern1024, testData->initial, sizeof pattern1024);
   memcpy (pattern2048, testData->initial, sizeof pattern2048);
   memcpy (pattern4096, testData->initial, sizeof pattern4096);
   memcpy (&mask32    , testData->mask32  , sizeof mask32    );
   memcpy (&mask64    , testData->mask64  , sizeof mask64    );
   memcpy (&mask128   , testData->mask128 , sizeof mask128   );
   memcpy (mask256    , testData->mask256 , sizeof mask256   );
   memcpy (mask512    , testData->mask512 , sizeof mask512   );
   memcpy (mask1024   , testData->mask1024, sizeof mask1024  );
   memcpy (mask2048   , testData->mask2048, sizeof mask2048  );
   memcpy (mask4096   , testData->mask4096, sizeof mask4096  );

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) pattern32 = lfsr32xmm (pattern32, mask32);

   printf ("lfsr32xmm   %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (&pattern32, testData->expected32, sizeof pattern32) != 0) printf ("-----lfsr32xmm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) pattern64 = lfsr64xmm (pattern64, mask64);

   printf ("lfsr64xmm   %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (&pattern64, testData->expected64, sizeof pattern64) != 0) printf ("-----lfsr64xmm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) pattern128 = lfsr128xmm (pattern128, mask128);

   printf ("lfsr128xmm  %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (&pattern128, testData->expected128, sizeof pattern128) != 0) printf ("-----lfsr128xmm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr256xmm (pattern256, mask256);

   printf ("lfsr256xmm  %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern256, testData->expected256, sizeof pattern256) != 0) printf ("-----lfsr256xmm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr512xmm (pattern512, mask512);

   printf ("lfsr512xmm  %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern512, testData->expected512, sizeof pattern512) != 0) printf ("-----lfsr512xmm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr1024xmm (pattern1024, mask1024);

   printf ("lfsr1024xmm %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern1024, testData->expected1024, sizeof pattern1024) != 0) printf ("-----lfsr1024xmm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr2048xmm (pattern2048, mask2048);

   printf ("lfsr2048xmm %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern2048, testData->expected2048, sizeof pattern2048) != 0) printf ("-----lfsr2048xmm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr4096xmm (pattern4096, mask4096);

   printf ("lfsr4096xmm %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern4096, testData->expected4096, sizeof pattern4096) != 0) printf ("-----lfsr4096xmm test failed-----\n");

   //----------------------------------------------------------------------------
   }
